;;;
;;; there is an example in test-func-env.lisp for (hopefully) every function
;;;

;;;
;;; Defines means and abstractions to use numerical functions as envelopes
;;; and provides a rich set of functions to generate, combine and transform
;;; these envelopes.
;;; 

#| ; TODO / TO THINK OF

|#

(in-package :fe)

;;;
;;; data structure
;;;

(defclass env ()
  ((env :accessor env 
	:initarg :env
	:initform (make-env #'identity)
	:type function
	:documentation "A unary numeric function, which defines an envelope in the function interval [0,1].")))

(defmethod env? ((x env))
  T)
(defmethod env? ((x T))
  NIL)

(defmethod y ((env env) (x number))
  "Access the y value of env at x."
  (funcall (env env) x))

(defmethod env->list ((env env) &optional (number 100))
  "Samples a env from 0 to 1 (including) and collects samples in a list. Number is number of samples. If number=1 the last env value is returned."
  (if (= number 1) (list (y env 1))
    (loop for n from 0 to 1 by (/ 1 (1- number))
	  collect (y env n))))
  
; (env->list (make-env #'sin :max pi) 10)


;;;
;;; generators
;;;

(defmethod make-env ((func function) &key (min 0) (max 1))
  "Builds a env from a given numeric function, such that interval [0,1] ofthe resulting env-func is mapped to [min, max] of the argument function."
  (make-instance 'env
    :env #'(lambda (x)
	     (assert (<= 0 x 1))
	     (funcall func (+ min (* x (- max min)))))))

; (env? (make-env #'identity :min 1 :max 2))
 
(defmethod make-env1 ((func function) &key (min 0) (max 1))
  "Builds a env from a given numeric function, such that interval [min,max] of resulting env is mapped to [0,1] of the argument function."
  (make-instance 'env
    :env #'(lambda (x)
	     (assert (<= min x max))
	     (funcall func (/ (-  x min) (- max min))))))

(flet ((aux (envs points)
	    ;; points: (0 ... 1)
	    (make-env
	     #'(lambda (x)
		 (let* ((pos (position x points :test #'<))
			(env (if pos
			  (make-env1 (env (nth (1- pos) envs))
					  :min (nth (1- pos) points)
					  :max (nth pos points))
			  (first (last envs)))))
		   (y env x))))))

  (defun env-seq (&rest funcenvs-and-points)
    "Combines an arbitrary number of envs to a single env. Expects its args of form &rest func num func num ... func. The numbers between the funcenvs specify the start resp. end point of a certain env. All numbers should be between 0--1 (exclusive)."
    (let ((points (append		; 0, <vals>, 1
		    (at-even-position
		     (cons 0 funcenvs-and-points)) (list 1))) 
	  (envs (at-even-position funcenvs-and-points)))
      (aux envs points)))
  
  (defun funcs->env (funcs &key (min 0) (max 1))
    "Converts a list of unary numeric functions to a single env. Max and min a given for all functions and the functions are equally spaced in the func env."
    (let ((points (loop for n from 0 to 1 by (/ 1 (length funcs))
			collect n))
	  (envs (mapcar #'(lambda (f)
				 (make-env f :min min :max max))
			     funcs)))
      ;;(format T "~%debug: points:~A envs ~A" points envs)
      (aux envs points)))

  (defmethod osciallator ((env env) (n number))
    "Defines a env by repeating env n times."
    (let ((envs (make-list n :initial-element env))
	  (points (loop for i from 0 to 1 by (/ 1 n)
			collect i)))
      (aux envs points))))

(defmethod saw ((n number) &key (amplitude 1) (offset 0))
  "Defines an env of saw shape (ascending) with n periods."
  (scale-env
   (osciallator (make-env #'(lambda (x)
				   (- (* 2 x) 1)))
		n)
   amplitude offset))
(defmethod saw1 ((n number) &key (amplitude 1) (offset 0))
  "Defines an env of saw shape (descending) with n periods."
  (reverse-env (saw n :amplitude amplitude :offset offset)))

(defmethod triangle ((n number) &key (amplitude 1) (offset 0))
  (scale-env
   (osciallator (linear-env (0 -1) (0.5 1) (1 -1)) n)
   amplitude offset))

;; ambitious function name!
(defmethod square ((n number) &key (amplitude 1) (offset 0))
  (scale-env
   (osciallator (make-env #'(lambda (x)
				   (if (< x 0.5)
				       -1
				     1)))
		n)
   amplitude offset))

;; irregular-steps (number max min)

(defun steps (&rest numbers)
  "Outputs a env composed of constant functions defined by numbers."
  (funcs->env (mapcar #'(lambda (x) #'(lambda (ignore) x))
			   numbers)))

(defun random-steps (n &key (min-y 0.0) (max-y 1.0))
  (assert (<= min-y max-y))
  (apply #'steps
	 (loop repeat n
	       collect (+ (random (abs (* (- max-y min-y)
					  1.0)))
			  min-y))))	       

(defmethod points->env ((func function) (points list))
  "Converts a list of points to a single env. A point is an x-y-pair as (xi yi). x values of the points range from 0--i (including), e.g., ((0 y1) (x2 y2) ... (1 yn)). The function defines the shape of the env segments and must return a env. It expects four numeric arguments, which describe the start and end points of the segment in the form (x1 y1 x2 y2)."
  (let* ((xs (mapcar #'first points))
	 (ys (mapcar #'second points))
	 (envs (mapcar func
		       (butlast xs)
		       (butlast ys)
		       (cdr xs)
		       (cdr ys))))
    (make-env
    #'(lambda (x)
	(let ((pos (position x (cdr xs) :test #'<=)))
	  (y (nth pos envs) x))))))

(defmethod rising-expon-env ((shape number))
  "Outputs an rising env with exponential shape. y in interval (0, 1]. The bigger shape is the steeper the env is. shape must be > 0."
  (assert (> shape 0))
  (make-env #'(lambda (x) (exp x)) :min (- shape) :max 0))

(defmethod constant-env ((value number))
  (make-env #'(lambda (x) value)))

;;;
;;; transformers
;;;

(defmethod reverse-env ((env env))
  (make-env
   #'(lambda (x)
       (y env (- 1 x)))))


(defmethod combine-envs ((combi-func function) &rest envs)
  "Returns a env which combines the given envs with an n-ary numeric function. envs can consist of envs and numeric values (i.e. constant functions) in any order. The combine-func must except as much numeric values as arguments to env are given."
  (make-env
   #'(lambda (x)
       (apply combi-func
	      (mapcar
	       #'(lambda (env)
		   (cond ((env? env) (y env x))
			 ((numberp env) env)
			 (T (error "~A is neither a env nor a number"
				   env))))
	       envs)))))


;(defun inverse-env (env &optional (axis (funcall env 0)))
;  )

(defmethod scale-env ((env env) (mul number) (add number))
  (combine-envs #'+ add
		(combine-envs #'* mul env)))


(defun rescale-env (env &key oldmin oldmax newmin newmax)
  ;; see CM 2.3.4 definition of rescale
  (combine-envs
   #'+
   (combine-envs #'* (combine-envs #'/ (combine-envs #'- newmax newmin)
				   (combine-envs #'- oldmax oldmin))
		 (combine-envs #'- env oldmin))
   newmin))

; (plot (rescale-env (make-env #'sin :max (* pi 2)) :oldmin -1 :oldmax 1 :newmin 0 :newmax 0.5))

;; noise...

;; hp-filter (env)
;; lp-filter (env)


(defmethod waveshape ((env1 env) (env2 env))
  "Returns an env which reads env1 'through' env2: the y value of env2 at a given x value access the y of env1 (take care to keep output of env2 in interval [0,1])."
  (make-env
   #'(lambda (x)
       (y env1 (y env2 x)))))
 


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; utils:
;;;

(defun at-even-position (in-list)
  (at-position in-list 2 0))


;;; EOF
